/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Andreas Krennmair <ak@synflood.at> wrote this file.  As long as you retain
 * this notice you can do whatever you want with this stuff. If we meet some day, 
 * and you think this stuff is worth it, you can buy me a beer in return.
 *   -- Andreas Krennmair <ak@synflood.at>
 * ----------------------------------------------------------------------------
 */
#include "poperattess.h"

#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include <buffer.h>
#include <case.h>
#include <scan.h>
#include <byte.h>
#include <str.h>
#include <stralloc.h>


enum { AUTHORIZATION = 1, TRANSACTION };

static void check_len_and_exit(int len)
{
	if (len < 0) {
		int e = errno;
		buffer_puts(buffer_2, "Error while reading: ");
		buffer_puts(buffer_2, strerror(e));
		buffer_putnlflush(buffer_2);
		exit(1);
	}
}

static void log_connection(char ip[4], uint16_t port)
{
	unsigned int i;
	buffer_puts(buffer_2, "Connection from ");
	for (i = 0; i < 4; i++) {
		buffer_putulong(buffer_2, (unsigned long) ip[i]);
		if (i < 3)
			buffer_puts(buffer_2, ".");
	}
	buffer_puts(buffer_2, ":");
	buffer_putulong(buffer_2, (unsigned long) port);
	buffer_putnlflush(buffer_2);
}

void handle_session(int fd, char ip[4], uint16_t port)
{
	char in_buf[BUFFER_INSIZE];
	char out_buf[BUFFER_OUTSIZE];
	int state = AUTHORIZATION;
	int cur_len;
	int exit_loop = 0;
	stralloc user, pass, cur_line;
	int have_user = 0, have_pass = 0;
	buffer in = BUFFER_INIT(read, fd, in_buf, sizeof(in_buf));
	buffer out = BUFFER_INIT(write, fd, out_buf, sizeof(out_buf));

	log_connection(ip, port);

	stralloc_init(&user);
	stralloc_init(&pass);
	stralloc_init(&cur_line);

	buffer_puts(&out, "+OK\r\n");
	buffer_flush(&out);

	while (!exit_loop) {
		cur_len = buffer_getnewline_sa(&in, &cur_line);

		check_len_and_exit(cur_len);

		stralloc_chomp(&cur_line);
		stralloc_0(&cur_line);

		switch (state) {
		case AUTHORIZATION:

			if (case_starts(cur_line.s, "user ")) {
				stralloc_copys(&user, cur_line.s + 5);
				have_user = 1;
			} else if (case_starts(cur_line.s, "pass ")) {
				stralloc_copys(&pass, cur_line.s + 5);
				have_pass = 1;
			} else if (case_starts(cur_line.s, "apop ")) {
				buffer_putsflush(&out, "-ERR APOP not supported\r\n");
				break;
			} else if (case_diffs(cur_line.s, "capa") == 0) {
				print_capa(&out);
				break;
			} else if (case_diffs(cur_line.s, "quit") == 0) {
				buffer_putsflush(&out, "+OK bye\r\n");
				exit_loop = 1;
				break;
			} else {
				buffer_putsflush(&out, "-ERR unknown command\r\n");
				break;
			}

			if (have_user && have_pass) {
				stralloc_0(&user);
				stralloc_0(&pass);
				if (auth_user(user.s, pass.s)) {
					state = TRANSACTION;
					if (init_transaction()) {
						buffer_putsflush(&out, "+OK authenticated\r\n");
					} else {
						buffer_putsflush(&out, "-ERR transaction initialization failed\r\n");
						exit_loop = 1;
					}
				} else {
					buffer_putsflush(&out, "-ERR authentication failed\r\n");
					exit_loop = 1;
				}
			} else {
				buffer_putsflush(&out, "+OK\r\n");
			}
			break;


		case TRANSACTION:

			if (case_diffs(cur_line.s, "stat") == 0) {
				print_stat(&out);
			} else if (case_diffs(cur_line.s, "capa") == 0) {
				print_capa(&out);
			} else if (case_starts(cur_line.s, "list")) {
				unsigned int msg_num;
				int rc;
				if (cur_line.s[4] == ' ') {
					rc = scan_uint(cur_line.s + 5, &msg_num);
					if (rc > 0) {
						list_msg(&out, msg_num);
					} else {
						buffer_putsflush(&out, "-ERR no such message\r\n");
					}
				} else {
					list_all(&out);
				}
			} else if (case_starts(cur_line.s, "retr ")) {
				unsigned int msg_num;
				int rc;
				rc = scan_uint(cur_line.s + 5, &msg_num);
				if (rc > 0) {
					retr_msg(&out, msg_num);
				} else {
					buffer_putsflush(&out, "-ERR invalid message number\r\n");
				}
			} else if (case_starts(cur_line.s, "dele ")) {
				unsigned int msg_num;
				int rc;
				rc = scan_uint(cur_line.s + 5, &msg_num);
				if (rc > 0) {
					dele_msg(&out, msg_num);
				} else {
					buffer_putsflush(&out, "-ERR invalid message number\r\n");
				}
			} else if (case_diffs(cur_line.s, "noop") == 0) {
				buffer_putsflush(&out, "+OK\r\n");
				/* do nothing else */
			} else if (case_diffs(cur_line.s, "rset") == 0) {
				do_reset();
				buffer_putsflush(&out, "+OK\r\n");
			} else if (case_diffs(cur_line.s, "quit") == 0) {
				/* this is the UPDATE state */
				update_maildrop();
				exit_loop = 1;
				buffer_putsflush(&out, "+OK bye\r\n");
			} else if (case_starts(cur_line.s, "top ")) {
				unsigned int msg_num, lines;
				int rc;
				rc = scan_uint(cur_line.s + 4, &msg_num);
				if (rc > 0) {
					rc = scan_uint(cur_line.s + 4 + rc + 1, &lines);
					if (rc > 0) {
						top_msg(&out, msg_num, lines);
					} else {
						buffer_putsflush(&out, "-ERR invalid number of lines\r\n");
					}
				} else {
					buffer_putsflush(&out, "-ERR invalid message number\r\n");
				}
			} else if (case_diffs(cur_line.s, "uidl") == 0) {
				uidl_all(&out);
			} else if (case_starts(cur_line.s, "uidl ")) {
				unsigned int msg_num;
				int rc;
				rc = scan_uint(cur_line.s + 5, &msg_num);
				if (rc > 0) {
					uidl_msg(&out, msg_num);
				} else {
					buffer_putsflush(&out, "-ERR invalid message number\r\n");
				}
			} else {
				buffer_putsflush(&out, "-ERR unknown command\r\n");
			}

			break;

		default:
			buffer_puts(buffer_2, "Bug: unknown state `");
			buffer_putulong(buffer_2, state);
			buffer_putsflush(buffer_2, "'\n");
			exit(1);
			break;
		}
	}
}
