/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * Andreas Krennmair <ak@synflood.at> wrote this file.  As long as you retain
 * this notice you can do whatever you want with this stuff. If we meet some day, 
 * and you think this stuff is worth it, you can buy me a beer in return.
 *   -- Andreas Krennmair <ak@synflood.at>
 * ----------------------------------------------------------------------------
 */
#include "poperattess.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <dirent.h>
#include <md5.h>

#include <str.h>
#include <stralloc.h>

struct msg {
	char *filename;
	struct stat stat_inf;
	int deleted;
};

static unsigned int number_mails;
static struct msg *mails;

static unsigned int count_mails(char *dir)
{
	unsigned int i = 0;
	DIR *d;
	d = opendir(dir);
	if (d) {
		struct dirent *de;
		while ((de = readdir(d)) != NULL) {
			if (str_diff(de->d_name, ".") != 0 && str_diff(de->d_name, "..") != 0) {
				++i;
			}
		}
		closedir(d);
	}
	return i;
}

static unsigned int fill_mail_array(struct msg *ma, char *dir, unsigned int off, unsigned int max)
{
	unsigned int i = off;
	DIR *d;
	struct dirent *de;
	d = opendir(dir);
	if (d) {
		for (i = off; i < max && (de = readdir(d));) {
			if (!str_equal(de->d_name, ".")
				&& !str_equal(de->d_name, "..")) {
				stralloc filename;
				stralloc_init(&filename);
				stralloc_cats(&filename, dir);
				stralloc_cats(&filename, "/");
				stralloc_cats(&filename, de->d_name);
				stralloc_0(&filename);
				if (0 == stat(filename.s, &(ma[i].stat_inf))
					&& S_ISREG(ma[i].stat_inf.st_mode)) {
					ma[i].filename = strdup(filename.s);
					ma[i].deleted = 0;
					++i;
				}
				stralloc_free(&filename);
			}
		}
		closedir(d);
	}
	return i;
}

int init_transaction(void)
{
	unsigned int count;
	if (chdir("/") != 0)
		return 0;
	number_mails = count_mails("cur") + count_mails("new");
	if (number_mails == 0)
		return 1;
	mails = malloc(number_mails * sizeof(struct msg));
	if (!mails)
		return 0;
	count = fill_mail_array(mails, "cur", 0, number_mails);
	number_mails = fill_mail_array(mails, "new", count, number_mails);
	return 1;
}

void print_stat(buffer * b)
{
	unsigned int i;
	unsigned int size = 0;
	unsigned int count = 0;
	for (i = 0; i < number_mails; ++i) {
		if (!mails[i].deleted) {
			size += mails[i].stat_inf.st_size;
			++count;
		}
	}
	buffer_puts(b, "+OK ");
	buffer_putulong(b, (unsigned long) count);
	buffer_puts(b, " ");
	buffer_putulong(b, (unsigned long) size);
	buffer_putsflush(b, "\r\n");
}

void list_msg(buffer * b, unsigned int n)
{
	if (n < 1 || n > number_mails || mails[n - 1].deleted) {
		buffer_putsflush(b, "-ERR no such message\r\n");
	} else {
		buffer_puts(b, "+OK ");
		buffer_putulong(b, (unsigned long) n);
		buffer_puts(b, " ");
		buffer_putulong(b, (unsigned long) (mails[n - 1].stat_inf.st_size));
		buffer_putsflush(b, "\r\n");
	}
}

void list_all(buffer * b)
{
	unsigned long i;
	unsigned long count = 0;
	for (i = 0; i < number_mails; ++i) {
		if (!mails[i].deleted)
			++count;
	}
	buffer_puts(b, "+OK ");
	buffer_putulong(b, count);
	buffer_puts(b, " messages\r\n");
	for (i = 0; i < number_mails; ++i) {
		if (!mails[i].deleted) {
			buffer_putulong(b, i + 1);
			buffer_puts(b, " ");
			buffer_putulong(b, (unsigned long) (mails[i].stat_inf.st_size));
			buffer_puts(b, "\r\n");
		}
	}
	buffer_putsflush(b, ".\r\n");
}

void retr_msg(buffer * b, unsigned int n)
{
	if (n < 1 || n > number_mails || mails[n - 1].deleted) {
		buffer_putsflush(b, "-ERR no such message\r\n");
	} else {
		char *f;
		off_t fs = mails[n - 1].stat_inf.st_size;
		int fd = open(mails[n - 1].filename, O_RDONLY);
		if (fd < 0) {
			buffer_putsflush(b, "-ERR message seems to have disappeared\r\n");	/* XXX */
			return;
		}
		f = mmap(NULL, fs, PROT_READ, MAP_PRIVATE, fd, 0);
		if (f == MAP_FAILED) {
			close(fd);
			buffer_putsflush(b, "-ERR cannot map message into memory\r\n");	/* XXX */
			return;
		}
		buffer_puts(b, "+OK message follows\r\n");
		buffer_put(b, f, fs);	/* TODO: escape lines with "." only */
		buffer_putsflush(b, ".\r\n");
		munmap(f, fs);
		close(fd);
	}
}

static void write_uid(buffer * b, unsigned int n)
{
	char *hash;
	if (n >= number_mails) {
		buffer_puts(b, "INVALID-MESSAGE-NUMBER-");
		buffer_putulong(b, (unsigned long) n);
		return;
	}
	/*
	 * create UID by (ab)using md5crypt with the basename 
	 * of the filename and the constant salt "poperattess" 
	 */
	hash = md5crypt(mails[n].filename + 4, "poperattess");	/* skip new/ or cur/ */
	buffer_puts(b, hash + 15);	/* skip $1$poperattess$ */
}

void top_msg(buffer * b, unsigned int n, unsigned int lines)
{
	char *f, *g, *mark;
	int fd;
	unsigned int i;
	off_t fs;
	if (n < 1 || n >= number_mails || mails[n - 1].deleted) {
		buffer_putsflush(b, "-ERR invalid message number\r\n");
		return;
	}
	fs = mails[n - 1].stat_inf.st_size;
	fd = open(mails[n - 1].filename, O_RDONLY);
	if (fd < 0) {
		buffer_putsflush(b, "-ERR message seems to have disappeared\r\n");
		return;
	}
	f = mmap(NULL, fs, PROT_READ, MAP_PRIVATE, fd, 0);
	if (f == MAP_FAILED) {
		close(fd);
		buffer_putsflush(b, "-ERR cannot map message into memory\r\n");
		return;
	}
	g = alloca(fs + 1);
	memcpy(g, f, fs);
	g[fs] = '\0';
	mark = strstr(g, "\n\n");
	if (!mark) {
		buffer_put(b, g, fs);
		buffer_putsflush(b, ".\r\n");
		return;
	} else {
		buffer_put(b, g, mark - g);
	}
	buffer_putsflush(b, "\r\n\r\n");
	if (lines > 0) {
		mark += 2;
	}
	for (i = 0; i < lines && mark != NULL; ++i) {
		char *new_mark = strstr(mark, "\n");
		if (!new_mark) {
			buffer_puts(b, mark);
		} else {
			new_mark += 1;		/* for the \n */
			buffer_put(b, mark, new_mark - mark);
		}
		mark = new_mark;
	}
	munmap(f, fs);
	close(fd);
	buffer_putsflush(b, ".\r\n");
}

void dele_msg(buffer * b, unsigned int n)
{
	if (n < 1 || n > number_mails || mails[n - 1].deleted) {
		buffer_putsflush(b, "-ERR no such message\r\n");
	} else {
		if (mails[n - 1].deleted) {
			buffer_putsflush(b, "-ERR message already deleted\r\n");
		} else {
			mails[n - 1].deleted = 1;
			buffer_puts(b, "+OK message ");
			buffer_putulong(b, (unsigned long) n);
			buffer_putsflush(b, " deleted\r\n");
		}
	}
}

void do_reset(void)
{
	unsigned int i;
	for (i = 0; i < number_mails; ++i) {
		mails[i].deleted = 0;
	}
}

void update_maildrop(void)
{
	unsigned int i;
	for (i = 0; i < number_mails; ++i) {
		if (mails[i].deleted) {
			unlink(mails[i].filename);
		}
	}
}

void uidl_msg(buffer * b, unsigned int msg_num)
{
	if (msg_num < 1 || msg_num > number_mails || mails[msg_num - 1].deleted) {
		buffer_putsflush(b, "-ERR invalid message number\r\n");
	} else {
		buffer_puts(b, "+OK ");
		buffer_putlong(b, (unsigned long) msg_num);
		buffer_puts(b, " ");
		write_uid(b, msg_num - 1);
		buffer_putsflush(b, "\r\n");
	}
}

void uidl_all(buffer * b)
{
	unsigned int i;
	buffer_puts(b, "+OK\r\n");
	for (i = 0; i < number_mails; ++i) {
		if (!mails[i].deleted) {
			buffer_putulong(b, (unsigned long) i + 1);
			buffer_puts(b, " ");
			write_uid(b, i);
			buffer_puts(b, "\r\n");
		}
	}
	buffer_putsflush(b, ".\r\n");
}

void print_capa(buffer * b) {
	buffer_puts(b, "+OK Capability list follows\r\n");
	buffer_puts(b, "TOP\r\n");
	buffer_puts(b, "USER\r\n");
	buffer_puts(b, "PIPELINING\r\n");
	buffer_puts(b, "UIDL\r\n");
	buffer_putsflush(b, ".\r\n");
}
